/*
 * Created on Sep 11, 2003
 */
package isac.bridge;

import isac.util.CalcElement;
import isac.util.ICalcIterator;

import java.io.Serializable;

/**
 * @author rgradisc
 */
public class CalcIterator implements ICalcIterator, Serializable {
	
  private CalcTree calcTree;
  private int calcTreeID;
  private int iteratorID;

  /**
  Construct a new calculation element iterator object attached 
  to the passed calculation tree. The iterator will
  reference the root element of the attached calculation tree 
  * @param calcTree calculation tree
  */
  protected CalcIterator(CalcTree calcTree) {
  	this.calcTree = calcTree;  	
  	this.calcTreeID = calcTree.getId();
  	calcTree.sendToBridge("Iterator "+calcTreeID+";");
  	//FIXXME
  	this.iteratorID = 1;  	 	
  }

  /**
  Reference the first (root) element in the attached calculation tree. 
  * @return Like all the moveXXX methods this returns true on success, 
  false if there is no element to move to. 
  */
  public boolean moveRoot() {
  	String s = calcTree.sendToBridge("moveRoot "+calcTreeID+" "+iteratorID+";");
    return true;
  }

  /**
  Reference the previous formula in the attached calculation tree.
  * @return Like all the moveXXX methods this returns true on success, 
  false if there is no element to move to. 
  */
  public boolean moveUp() {
	String s = calcTree.sendToBridge("moveUp "+calcTreeID+" "+iteratorID+";");
    return true;
  }

  /**
  Reference the next formula in the attached calculation tree.
  * @return Like all the moveXXX methods this returns true on success, 
  false if there is no element to move to. 
  */
  public boolean moveDown() {
	String s = calcTree.sendToBridge("moveDown "+calcTreeID+" "+iteratorID+";");
    return true;
  }

  /**
  Enter the next coarser level of refinement in the attached calculation tree.
  * @return Like all the moveXXX methods this returns true on success, 
  false if there is no element to move to. 
  */
  public boolean moveLevelUp() {
	String s = calcTree.sendToBridge("moveLevelUp "+calcTreeID+" "+iteratorID+";");
    return true;
  }

  /**
  Enter the next more detailed level of refinement in the 
  attached calculation tree.
  * @return Like all the moveXXX methods this returns true on success, 
  false if there is no element to move to. 
  */
  public boolean moveLevelDown() {
	String s = calcTree.sendToBridge("moveLevelDown "+calcTreeID+" "+iteratorID+";");
    return true;
  }

  /**
  Reference the tactic which resulted in the currently selected element of 
  the attached calculation tree. 
  <p>A special case arises if the tactic to continue the current calculation 
  is solving a subproblem. The subproblem is a calculation on its own, 
  represented by a separate calculation(sub-)tree object. A reference to the 
  calculationtree representing the subproblem is passed to the worksheet. 
  The worksheet will handle the situation accordingly, e.g. by opening a new 
  worksheet or displaying the new calculationtree (indented) in the current
  worksheet. 
  * @return Like all the moveXXX methods this returns true on success, 
  false if there is no element to move to. 
  */
  public boolean moveTactic() {
	String s = calcTree.sendToBridge("moveTactic "+calcTreeID+" "+iteratorID+";");
    return true;
  }

  /**
  Reference the formula which resulted in the currently selected element 
  of the attached calculation tree.
  */
  public boolean moveFormula() {
	String s = calcTree.sendToBridge("moveFormula "+calcTreeID+" "+iteratorID+";");
  	return true;
  }

  /* 
  public Formula getFormula() {
    return new Formula();
  }

  public Tactic getTactic() {
    return new Tactic();
  }
  */
   
  /**
  The current level of nesting into subcalculations.
  * @return An int representing level of nesting (higher == more detailed)
  <p>Special values:
  <p>0: outmost level: only first and last step of calculation are on this level 
  */ 
  public int getLevel() {
    return 0;
  }

  /* (non-Javadoc)
   * @see isac.wsdialog.ICalcIterator#getElement()
   */
  public CalcElement getElement() {
    // TODO Auto-generated method stub
    return null;
  }

  /* (non-Javadoc)
   * @see isac.wsdialog.ICalcIterator#clone()
   */
  public Object clone() {    
    return null;
  }

  /* (non-Javadoc)
   * @see java.lang.Comparable#compareTo(java.lang.Object)
   */
  public int compareTo(Object o) {  	    
    return 0;
  }
}
