/*
 * Created on Aug 21, 2003
 */
package isac.bridge;

/**
 * 
 * @author rgradisc
 *
 * Stores times of requests to sml-kernel
 * and informs Clients2KIServer to kill sml-kernel if it sends no response for too long 
 */

import java.util.Vector;

public class TimeCheckerThread extends Thread {

  private long waitMillis;

  private Vector timeOutTimes;
  private BridgeMain bridge;  

  public TimeCheckerThread(BridgeMain bridge, long waitMillis) {
    this.bridge = bridge;
    this.timeOutTimes = bridge.getTimeOutTimes();  
    this.waitMillis = waitMillis;
  }

  public void run() {
    try {
      while (true) {
        if (!timeOutTimes.isEmpty()) {
          long first = ((Long) timeOutTimes.firstElement()).longValue();
          long timeOut = first + waitMillis;
          long now = System.currentTimeMillis();
          if (now > timeOut) {
            // timeOut: no response from sml-kernel within TIME_OUT: restart sml-kernel
            bridge.restartSML();
          } else {
            // sleep until first time in vector would run out (plus a little buffer)
            long sleepTime = timeOut - now + 20;
            bridge.log(
              1,
              "TimeCheckerThread: "
                + timeOutTimes.size()
                + " request(s) waiting for response, sleeping for "
                + sleepTime
                + " millis");
            sleep(sleepTime);
          }
        } else {
          // TimesVector is empty
          bridge.log(
            1,
            "TimeCheckerThread: no requests, sleeping for "
              + waitMillis
              + " millis");
          sleep(waitMillis);
        }
      }
    } catch (InterruptedException e) {
      bridge.log(1, "TimeCheckerThread interrupted");
    }
  }

}
