/*
 * Created on Aug 8, 2003
*/
package isac.bridge;

/**
 * @author rgradisc
*/

import java.io.FileOutputStream;
import java.io.StringReader;
import java.io.StringWriter;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

public class TransformXSLT {

  private TransformerFactory tFactory;

  public TransformXSLT() {
    tFactory = TransformerFactory.newInstance();
  }

  public void transform(String xmlFile, String xslFile, String outFile) {
    // Use the static TransformerFactory.newInstance() method to instantiate 
    // a TransformerFactory. The javax.xml.transform.TransformerFactory 
    // system property setting determines the actual class to instantiate --
    // org.apache.xalan.transformer.TransformerImpl. 

    // Use the TransformerFactory to instantiate a Transformer that will work with  
    // the stylesheet you specify. This method call also processes the stylesheet
    // into a compiled Templates object. 
    try {
      Transformer transformer =
        tFactory.newTransformer(new StreamSource(xslFile));
      transformer.transform(
        new StreamSource(xmlFile),
        new StreamResult(new FileOutputStream(outFile)));

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public String stringTransform(String xml, String xsl) {
    StringWriter stringWriter = new StringWriter();
    try {
      Transformer transformer =
        tFactory.newTransformer(new StreamSource(new StringReader(xsl)));
      transformer.transform(
        new StreamSource(new StringReader(xml)),
        new StreamResult(stringWriter));
    } catch (Exception e) {
      e.printStackTrace();
    }
    return stringWriter.toString();
  }

  public static void main(String[] args) {
    String path = "/radish/hda-local/isac/data/text";
    TransformXSLT t = new TransformXSLT();
    System.out.println("Started transforming");
    long before = System.currentTimeMillis();
    t.transform(path + ".xml", path + ".xsl", path + "out.xml");
    long after = System.currentTimeMillis();
    System.out.println(
      "Done file transforming: " + (after - before) + " millis needed");
    //result should be: 1 (Add_Given "equality (1 + -1 * 2 + x = 0)")

    before = System.currentTimeMillis();
    String xmlString = "<?xml version=\"1.0\"?><doc>Hello</doc>";
    String xslString =
      "<?xml version=\"1.0\"?>"
        + "<xsl:stylesheet xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" version=\"1.0\">"
        + "<xsl:template match=\"doc\">"
        + "<out><xsl:value-of select=\".\"/></out>"
        + "</xsl:template>"
        + "</xsl:stylesheet>";
    System.out.println(
      "StringTransform:" + t.stringTransform(xmlString, xslString));
    after = System.currentTimeMillis();
    System.out.println(
      "Done string transforming: " + (after - before) + " millis needed");
  }
}