(* use"pbl_met.sml";
   use"xmlsrc/pbl-met-hierarchy.sml";
   use"../xmlsrc/pbl-met-hierarchy.sml";
   *)

type path = string;
type filename = string;

fun str2file (fnm:filename) (str:string) =
    let val file = TextIO.openOut fnm
    in (TextIO.output (file, str);
	TextIO.flushOut file;
	TextIO.closeOut file) end;
fun pos2filename [] = raise error "pos2filename called with []"
  | pos2filename [i] = "_" ^ string_of_int i ^ ".xml"
  | pos2filename (i::is) = "_" ^ string_of_int i ^ pos2filename is;
(* pos2filename [1,22,3];
val it = "_1_22_3.xml" : string
*)
fun id2filename [] = raise error "id2filename called with []"
  | id2filename [s] = s ^ ".xml"
  | id2filename (s::ss) = s ^ "_" ^ id2filename ss;
(* id2filename ["linear","univariate","equation"];
val it = "linear_univariate_equation.xml" : string
*)



(*ad DTD: a NODE contains an ID and zero or more NODEs*)
fun hierarchy pm(*"pbl" | "met"*) h =
    let val j = indentation
	fun nd i p (Ptyp (id,_,ns)) = 
	    let val p' = lev_on p
	    in
		(indt i) ^ "<NODE>\n" ^ 
		(indt (i+j)) ^ "<ID> " ^ id ^ " </ID>\n" ^ 
		(indt (i+j)) ^ 
		        "<CONTENTREF> " ^ pm ^ pos2filename p' ^ 
			" </CONTENTREF>\n" ^
		(nds (i+j) (lev_dn p') ns) ^ 
		(indt i) ^ "</NODE>\n"
	    end
	and nds _ _ [] = ""
	  | nds i p (n::ns) = (nd i p n) ^ (nds i (lev_on p) ns);
    in nds 0 [0] h end;
fun hierarchy pm(*"pbl" | "met"*) h =
    let val j = indentation
	fun nd i p (Ptyp (id,_,ns)) = 
	    let val p' = lev_on p
	    in
		(indt i) ^ "<NODE>\n" ^ 
		(indt (i+j)) ^ "<ID> " ^ id ^ " </ID>\n" ^ 
		(indt (i+j)) ^ 
		               "<NO> " ^ 
			       (string_of_int o last_elem) p' ^ " </NO>\n" ^ 
		(indt (i+j)) ^ 
		        "<CONTENTREF> " ^ pm ^ pos2filename p' ^ 
			" </CONTENTREF>\n" ^
		(nds (i+j) (lev_dn p') ns) ^ 
		(indt i) ^ "</NODE>\n"
	    end
	and nds _ _ [] = ""
	  | nds i p (n::ns) = (nd i p n) ^ (nds i (lev_on p) ns);
    in nds 0 [0] h end;
(* (writeln o hierarchy) (!ptyps);
   *)

fun pbl_hierarchy2file (path:path) = 
    str2file (path ^ "pbl_hierarchy.xml") 
	     ("<PBLHIERARCHY>\n" ^
	     (hierarchy "pbl" (!ptyps)) ^
	     "</PBLHIERARCHY>");
fun met_hierarchy2file (path:path) = 
    str2file (path ^ "met_hierarchy.xml") 
	     ("<METHIERARCHY>\n" ^
	     (hierarchy "met" (!mets)) ^
	     "<METHIERARCHY>");
(* pbl_hierarchy2file "../xml/pbl/";
   met_hierarchy2file "../xml/met/";
   *)


val i = indentation;

fun pbl2xml (id:pblRD) ({cas,met,ppc,prls,thy,where_}:pbt) = 
    "<NODECONTENT>\n" 
    ^
    (((id2xml i) o rev) id) 
    ^ 
    indt i ^ "<META> </META>\n" ^
    indt i ^ "<EXPLANATIONS> </EXPLANATIONS>\n" ^
    pattern2xml i ppc where_
    ^
    (indt i) ^ "<THEORY> " ^ theory2theory' thy ^ " </THEORY>\n"
    ^
    (case met of [] => (indt i) ^ "<METHODS> </METHODS>\n"
	       | _ => (indt i) ^ "<METHODS>\n" ^
		      foldl op^ ("", map (id2xml (i+i)) met) ^
		      (indt i) ^ "</METHODS>\n")
    ^
    (case cas of None => (indt i) ^ "<CAS> </CAS>\n"
	       | Some t => (indt i) ^ "<CAS>\n" ^ term2xml i t ^ "\n" ^
			   (indt i) ^ "</CAS>\n")    
    ^
    (indt i) ^ "<EVALPRECOND> " ^ (#id o rep_rls) prls ^ " </EVALPRECOND>\n"
    ^
    "</NODECONTENT>";

fun scr2xml j EmptyScr = 
    (indt j ^ "<SCRIPT> empty_script </SCRIPT>\n"):xml
  | scr2xml j (Script t) =
    indt j ^ "<SCRIPT>\n" ^ 
    term2xml j (inst_abs (assoc_thy "Isac.thy") t) ^ "\n" ^
    indt j ^ "</SCRIPT>\n"
  | scr2xml j (Rfuns _) =
    indt j ^ "<SCRIPT> 'reverse rewriting' (functions not displayed) \
	     \</SCRIPT>\n";
fun calc2xml j ((scr_op, (isa_op, _)):calc) =
    indt i ^ "<CALCULATE> (" ^ scr_op ^ ", (" ^ isa_op ^ ")) </CALCULATE>\n";
fun calcs2xml j [] = ("":xml)
  | calcs2xml j (r::rs) = calc2xml j r ^ calcs2xml j rs;

fun met2xml (id:metID) 
	    ({ppc,pre,scr,calc,crls,erls,nrls,prls,srls,rew_ord'}:met) = 
    "<NODECONTENT>\n" ^
    id2xml i id ^ 
    indt i ^ "<META> </META>\n" ^
    indt i ^ "<EXPLANATIONS> </EXPLANATIONS>\n" ^
    pattern2xml i ppc pre ^
    indt i ^ "<EVALPRECOND> " ^ (#id o rep_rls) prls ^ " </EVALPRECOND>\n" ^
    indt i ^ "<EVALCOND> "    ^ (#id o rep_rls) erls ^ " </EVALCOND>\n" ^
    indt i ^ "<EVALLISTEXPR> "^ (#id o rep_rls) srls ^ " </EVALLISTEXPR>\n" ^
    indt i ^ "<CHECKELEMENTWISE> " ^ 
                                (#id o rep_rls) crls ^ 
				" </CHECKELEMENTWISE>\n" ^
    indt i ^ "<NORMALFORM> "  ^ (#id o rep_rls) nrls ^ " </NORMALFORM>\n" ^
    indt i ^ "<REWORDER> " ^ rew_ord' ^ " </REWORDER>\n" ^
    calcs2xml i calc ^
    scr2xml i scr ^
    "</NODECONTENT>";

fun pbl2file path (pos:pos) id pbl =
    (writeln ("### pbl2file: id = " ^ strs2str id);
    ((str2file (path ^ "pbl" ^ pos2filename pos)) o (pbl2xml id)) pbl
    );
fun met2file path (pos:pos) id met =
    ((str2file (path ^ "met" ^ pos2filename pos)) o (met2xml id)) met;

fun nd (pa:path) ids po wfn (Ptyp (id,[n],ns)) = 
    let val po' = lev_on po
    in wfn pa po' (ids@[id]) n; 
    nds pa (ids@[id]) ((lev_dn po'):pos) wfn ns end
and nds _ _ _ _ [] = ()
  | nds pa ids po wfn (n::ns) = (nd pa ids po wfn n;  
				 nds pa ids (lev_on po) wfn ns);


fun pbls2file (p:path) = nds p [] [0] pbl2file (!ptyps);
fun mets2file (p:path) = nds p [] [0] met2file (!mets);
(* pbls2file "../xml/pbl/";
   mets2file "../xml/met/";
   *)